def create_webserver(server_config, backend="aiohttp"):
    if backend == "aiohttp":
        from .aiohttp import create_webserver

        return create_webserver(server_config)

    if backend == "generic":
        from .generic import create_webserver

        return create_webserver(server_config)

    if backend == "tornado":
        from .tornado import create_webserver

        return create_webserver(server_config)

    if backend == "jupyter":
        from .jupyter import create_webserver

        return create_webserver(server_config)

    raise Exception(f"{backend} backend is not implemented")


def launcher_start(args, config, backend="aiohttp"):
    if backend == "aiohttp":
        from .aiohttp.launcher import startWebServer

        return startWebServer(args, config)

    if backend == "generic":
        from .generic import startWebServer

        return startWebServer(args, config)

    if backend == "tornado":
        from .tornado import startWebServer

        return startWebServer(args, config)

    if backend == "jupyter":
        from .jupyter import startWebServer

        return startWebServer(args, config)

    raise Exception(f"{backend} backend is not implemented")



import os
import sys
import secrets
import msgpack
from typing import Dict, Tuple, Union

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict

UINT32_LENGTH = 4
ID_LOCATION = 0
ID_LENGTH = UINT32_LENGTH
MESSAGE_OFFSET_LOCATION = ID_LOCATION + ID_LENGTH
MESSAGE_OFFSET_LENGTH = UINT32_LENGTH
MESSAGE_SIZE_LOCATION = MESSAGE_OFFSET_LOCATION + MESSAGE_OFFSET_LENGTH
MESSAGE_SIZE_LENGTH = UINT32_LENGTH

HEADER_LENGTH = ID_LENGTH + MESSAGE_OFFSET_LENGTH + MESSAGE_SIZE_LENGTH


def _encode_header(id: int, offset: int, size: int) -> bytes:
    return (
        id.to_bytes(ID_LENGTH, "little", signed=False)
        + offset.to_bytes(MESSAGE_OFFSET_LENGTH, "little", signed=False)
        + size.to_bytes(MESSAGE_SIZE_LENGTH, "little", signed=False)
    )


def _decode_header(header: bytes) -> Tuple[int, int, int]:
    id = int.from_bytes(
        header[ID_LOCATION:ID_LENGTH],
        "little",
        signed=False,
    )
    offset = int.from_bytes(
        header[
            MESSAGE_OFFSET_LOCATION : MESSAGE_OFFSET_LOCATION + MESSAGE_OFFSET_LENGTH
        ],
        "little",
        signed=False,
    )
    size = int.from_bytes(
        header[MESSAGE_SIZE_LOCATION : MESSAGE_SIZE_LOCATION + MESSAGE_SIZE_LENGTH],
        "little",
        signed=False,
    )
    return id, offset, size


def generate_chunks(message: bytes, max_size: int):
    total_size = len(message)

    if max_size == 0:
        max_content_size = total_size
    else:
        max_content_size = max(max_size - HEADER_LENGTH, 1)

    id = int.from_bytes(secrets.token_bytes(ID_LENGTH), "little", signed=False)

    offset = 0

    while offset < total_size:
        header = _encode_header(id, offset, total_size)
        chunk_content = message[offset : offset + max_content_size]

        yield header + chunk_content

        offset += max_content_size

    return


class PendingMessage(TypedDict):
    received_size: int
    content: bytearray


# This un-chunker is vulnerable to DOS.
# If it receives a message with a header claiming a large incoming message
# it will allocate the memory blindly even without actually receiving the content
# Chunks for a given message can come in any order
# Chunks across messages can be interleaved.
class UnChunker:
    pending_messages: Dict[bytes, PendingMessage]
    max_message_size: int

    def __init__(self):
        self.pending_messages = {}
        self.max_message_size = int(os.environ.get("WSLINK_AUTH_MSG_SIZE", 512))

    def set_max_message_size(self, size):
        self.max_message_size = size

    def release_pending_messages(self):
        self.pending_messages = {}

    def process_chunk(self, chunk: bytes) -> Union[bytes, None]:
        header, chunk_content = chunk[:HEADER_LENGTH], chunk[HEADER_LENGTH:]
        id, offset, total_size = _decode_header(header)

        pending_message = self.pending_messages.get(id, None)

        if pending_message is None:
            if total_size > self.max_message_size:
                raise ValueError(
                    f"""Total size for message {id} exceeds the allocation limit allowed.
    Maximum size = {self.max_message_size},
    Received size = {total_size}."""
                )

            pending_message = PendingMessage(
                received_size=0, content=bytearray(total_size)
            )
            self.pending_messages[id] = pending_message

        # This should never happen, but still check it
        if total_size != len(pending_message["content"]):
            del self.pending_messages[id]
            raise ValueError(
                f"Total size in chunk header for message {id} does not match total size declared by previous chunk."
            )

        content_size = len(chunk_content)
        content_view = memoryview(pending_message["content"])
        content_view[offset : offset + content_size] = chunk_content
        pending_message["received_size"] += content_size

        if pending_message["received_size"] >= total_size:
            full_message = pending_message["content"]
            del self.pending_messages[id]
            return msgpack.unpackb(bytes(full_message))

        return None


class StreamPendingMessage(TypedDict):
    received_size: int
    total_size: int
    unpacker: msgpack.Unpacker


# This un-chunker is more memory efficient
# (each chunk is passed immediately to msgpack)
# and it will only allocate memory when it receives content.
# Chunks for a given message are expected to come sequentially
# Chunks across messages can be interleaved.
class StreamUnChunker:
    pending_messages: Dict[bytes, StreamPendingMessage]

    def __init__(self):
        self.pending_messages = {}

    def set_max_message_size(self, _size):
        pass

    def release_pending_messages(self):
        self.pending_messages = {}

    def process_chunk(self, chunk: bytes) -> Union[bytes, None]:
        header, chunk_content = chunk[:HEADER_LENGTH], chunk[HEADER_LENGTH:]
        id, offset, total_size = _decode_header(header)

        pending_message = self.pending_messages.get(id, None)

        if pending_message is None:
            pending_message = StreamPendingMessage(
                received_size=0,
                total_size=total_size,
                unpacker=msgpack.Unpacker(max_buffer_size=total_size),
            )
            self.pending_messages[id] = pending_message

        # This should never happen, but still check it
        if offset != pending_message["received_size"]:
            del self.pending_messages[id]
            raise ValueError(
                f"""Received an unexpected chunk for message {id}.
    Expected offset = {pending_message['received_size']},
    Received offset = {offset}."""
            )

        # This should never happen, but still check it
        if total_size != pending_message["total_size"]:
            del self.pending_messages[id]
            raise ValueError(
                f"""Received an unexpected total size in chunk header for message {id}.
    Expected size = {pending_message['total_size']},
    Received size = {total_size}."""
            )

        content_size = len(chunk_content)
        pending_message["received_size"] += content_size

        unpacker = pending_message["unpacker"]
        unpacker.feed(chunk_content)

        full_message = None

        try:
            full_message = unpacker.unpack()
        except msgpack.OutOfData:
            pass  # message is incomplete, keep ingesting chunks

        if full_message is not None:
            del self.pending_messages[id]

            if pending_message["received_size"] < total_size:
                # In principle feeding a stream to the unpacker could yield multiple outputs
                # for example unpacker.feed(b'0123') would yield b'0', b'1', ect
                # or concatenated packed payloads would yield two or more unpacked objects
                # but in our use case we expect a full message to be mapped to a single object
                raise ValueError(
                    f"""Received a parsable payload shorter than expected for message {id}.
    Expected size = {total_size},
    Received size = {pending_message['received_size']}."""
                )

        return full_message



import argparse
import asyncio
import datetime
import io
import json
import logging
import os
import re
import string
import subprocess
import sys
import time
import uuid

from random import choice

from wslink import backends

logger = logging.getLogger(__name__)

STATUS_OK = 200
STATUS_BAD_REQUEST = 400
STATUS_NOT_FOUND = 404
STATUS_SERVICE_UNAVAILABLE = 503

sample_config_file = """
Here is a sample of what a configuration file could look like:

    {
        // ===============================
        // General launcher configuration
        // ===============================

        "configuration": {
            "host" : "localhost",
            "port" : 8080,
            "endpoint": "paraview",                   // SessionManager Endpoint
            "content": "/.../www",                    // Optional: Directory shared over HTTP
            "proxy_file" : "/.../proxy-mapping.txt",  // Proxy-Mapping file for Apache
            "sessionURL" : "ws://${host}:${port}/ws", // ws url used by the client to connect to the started process
            "timeout" : 25,                           // Wait time in second after process start
            "log_dir" : "/.../viz-logs",              // Directory for log files
            "fields" : ["file", "host", "port"]       // List of fields that should be send back to client
                                                             // include "secret" if you provide it as an --authKey to the app
            "sanitize": {                             // Check information coming from the client
                "cmd": {
                    "type": "inList",                 // 'cmd' must be one of the strings in 'list'
                    "list": [
                        "me", "you", "something/else/altogether", "nothing-to-do"
                    ],
                    "default": "nothing-to-do"        // If the string doesn't match, replace it with the default.
                                                      // Include the default in your list
                },
                "cmd2": {                             // 'cmd2' must match the regexp provided, example: not a quote
                    "type": "regexp",
                    "regexp": "^[^\"]*$",             // Make sure to include '^' and '$' to match the entire string!
                    "default": "nothing"
                }
            }
        },

        // ===============================
        // Useful session vars for client
        // ===============================

        "sessionData" : { "key": "value" },      // Dictionary of values interesting to the client

        // ===============================
        // Resources list for applications
        // ===============================

        "resources" : [ { "host" : "localhost", "port_range" : [9001, 9003] } ],

        // ===============================
        // Set of properties for cmd line
        // ===============================

        "properties" : {
            "vtkpython" : "/.../VTK/build/bin/vtkpython",
            "pvpython" : "/.../ParaView/build/bin/pvpython",
            "vtk_python_path": "/.../VTK/build/Wrapping/Python/vtk/web",
            "pv_python_path": "/.../ParaView/build/lib/site-packages/paraview/web",
            "plugins_path": "/.../ParaView/build/lib",
            "dataDir": "/.../path/to/data/directory"
        },

        // ===============================
        // Application list with cmd lines
        // ===============================

        "apps" : {
            "cone" : {
                "cmd" : [
                    "${vtkpython}", "${vtk_python_path}/vtk_web_cone.py", "--port", "$port" ],
                "ready_line" : "Starting factory"
            },
            "graph" : {
                "cmd" : [
                    "${vtkpython}", "${vtk_python_path}/vtk_web_graph.py", "--port", "$port",
                    "--vertices", "${numberOfVertices}", "--edges", "${numberOfEdges}" ],
                "ready_line" : "Starting factory"
            },
            "phylotree" : {
                "cmd" : [
                    "${vtkpython}", "${vtk_python_path}/vtk_web_phylogenetic_tree.py", "--port", "$port",
                    "--tree", "${dataDir}/visomics/${treeFile}", "--table", "${dataDir}/visomics/${tableFile}" ],
                "ready_line" : "Starting factory"
            },
            "filebrowser" : {
                "cmd" : [
                    "${vtkpython}", "${vtk_python_path}/vtk_web_filebrowser.py",
                    "--port", "${port}", "--data-dir", "${dataDir}" ],
                "ready_line" : "Starting factory"
            },
            "data_prober": {
                "cmd": [
                    "${pvpython}", "-dr", "${pv_python_path}/pv_web_data_prober.py",
                    "--port", "${port}", "--data-dir", "${dataDir}", "-f" ],
                "ready_line" : "Starting factory"
            },
            "visualizer": {
                "cmd": [
                    "${pvpython}", "-dr", "${pv_python_path}/pv_web_visualizer.py",
                    "--plugins", "${plugins_path}/libPointSprite_Plugin.so", "--port", "${port}",
                    "--data-dir", "${dataDir}", "--load-file", "${dataDir}/${fileToLoad}",
                    "--authKey", "${secret}", "-f" ],  // Use of ${secret} means it needs to be provided to the client, in "fields", above.
                "ready_line" : "Starting factory"
            },
            "loader": {
                "cmd": [
                    "${pvpython}", "-dr", "${pv_python_path}/pv_web_file_loader.py",
                    "--port", "${port}", "--data-dir", "${dataDir}",
                    "--load-file", "${dataDir}/${fileToLoad}", "-f" ],
                "ready_line" : "Starting factory"
            },
            "launcher" : {
                "cmd": [
                    "/.../ParaView/Web/Applications/Parallel/server/launcher.sh",
                    "${port}", "${client}", "${resources}", "${file}" ],
                "ready_line" : "Starting factory"
            },
            "your_app": {
                "cmd": [
                    "your_shell_script.sh", "--resource-host", "${host}", "--resource-port", "${port}",
                    "--session-id", "${id}", "--generated-password", "${secret}",
                    "--application-key", "${application}" ],
                "ready_line": "Output line from your shell script indicating process is ready"
        }
    }
"""

# =============================================================================
# Helper module methods
# =============================================================================


def remove_comments(json_like):
    """
    Removes C-style comments from *json_like* and returns the result.  Example::
        >>> test_json = '''\
        {
            "foo": "bar", // This is a single-line comment
            "baz": "blah" /* Multi-line
            Comment */
        }'''
        >>> remove_comments('{"foo":"bar","baz":"blah",}')
        '{\n    "foo":"bar",\n    "baz":"blah"\n}'

        From: https://gist.github.com/liftoff/ee7b81659673eca23cd9fc0d8b8e68b7
    """
    comments_re = re.compile(
        r'//.*?$|/\*.*?\*/|\'(?:\\.|[^\\\'])*\'|"(?:\\.|[^\\"])*"',
        re.DOTALL | re.MULTILINE,
    )

    def replacer(match):
        s = match.group(0)
        if s[0] == "/":
            return ""
        return s

    return comments_re.sub(replacer, json_like)


def generatePassword():
    return "".join(choice(string.ascii_letters + string.digits) for _ in range(16))


# -----------------------------------------------------------------------------


def validateKeySet(obj, expected_keys, object_name):
    all_key_found = True
    for key in expected_keys:
        if not key in obj:
            print("ERROR: %s is missing %s key." % (object_name, key))
            all_key_found = False
    return all_key_found


def checkSanitize(key_pair, sanitize):
    if not sanitize:
        return
    for key in sanitize:
        if key in key_pair:
            checkItem = sanitize[key]
            value = key_pair[key]
            if checkItem["type"] == "inList":
                if not value in checkItem["list"]:
                    logger.warning(
                        "key %s: sanitize %s with default" % (key, key_pair[key])
                    )
                    key_pair[key] = checkItem["default"]
            elif checkItem["type"] == "regexp":
                if not "compiled" in checkItem:
                    # User is responsible to add begin- and end- string symbols, to make sure entire string is matched.
                    checkItem["compiled"] = re.compile(checkItem["regexp"])
                if checkItem["compiled"].match(value) == None:
                    logger.warning(
                        "key %s: sanitize %s with default" % (key, key_pair[key])
                    )
                    key_pair[key] = checkItem["default"]


# -----------------------------------------------------------------------------
# guard against malicious clients - make sure substitution is expected, if 'sanitize' is provided
# -----------------------------------------------------------------------------


def replaceVariables(template_str, variable_list, sanitize):
    for key_pair in variable_list:
        checkSanitize(key_pair, sanitize)
        item_template = string.Template(template_str)
        template_str = item_template.safe_substitute(key_pair)

    if "$" in template_str:
        logger.error("Some properties could not be resolved: " + template_str)

    return template_str


# -----------------------------------------------------------------------------


def replaceList(template_list, variable_list, sanitize):
    result_list = []
    for template_str in template_list:
        result_list.append(replaceVariables(template_str, variable_list, sanitize))
    return result_list


# -----------------------------------------------------------------------------


def filterResponse(obj, public_keys):
    public_keys.extend(["id", "sessionURL", "sessionManagerURL"])
    filtered_output = {}
    for field in obj:
        if field in public_keys:
            filtered_output[field] = obj[field]
    return filtered_output


# -----------------------------------------------------------------------------


def extractSessionId(request):
    path = request.path.split("/")
    if len(path) < 3:
        return None
    return str(path[2])


def jsonResponse(payload):
    return json.dumps(payload, ensure_ascii=False).encode("utf8")


# =============================================================================
# Session manager
# =============================================================================


class SessionManager(object):
    def __init__(self, config, mapping):
        self.sessions = {}
        self.config = config
        self.resources = ResourceManager(config["resources"])
        self.mapping = mapping
        self.sanitize = config["configuration"]["sanitize"]

    def createSession(self, options):
        # Assign id and store options
        id = str(uuid.uuid1())

        # Assign resource to session
        host, port = self.resources.getNextResource()

        # Do we have resources
        if host:
            options["id"] = id
            options["host"] = host
            options["port"] = port
            if not "secret" in options:
                options["secret"] = generatePassword()
            options["sessionURL"] = replaceVariables(
                self.config["configuration"]["sessionURL"],
                [options, self.config["properties"]],
                self.sanitize,
            )
            options["cmd"] = replaceList(
                self.config["apps"][options["application"]]["cmd"],
                [options, self.config["properties"]],
                self.sanitize,
            )

            if "sessionData" in self.config:
                for key in self.config["sessionData"]:
                    options[key] = replaceVariables(
                        self.config["sessionData"][key],
                        [options, self.config["properties"]],
                        self.sanitize,
                    )

            self.sessions[id] = options
            self.mapping.update(self.sessions)
            return options

        return None

    def deleteSession(self, id):
        host = self.sessions[id]["host"]
        port = self.sessions[id]["port"]
        self.resources.freeResource(host, port)
        del self.sessions[id]
        self.mapping.update(self.sessions)

    def getSession(self, id):
        if id in self.sessions:
            return self.sessions[id]
        return None


# =============================================================================
# Proxy manager
# =============================================================================


class ProxyMappingManager(object):
    def update(sessions):
        pass


class ProxyMappingManagerTXT(ProxyMappingManager):
    def __init__(self, file_path, pattern="%s %s:%d\n"):
        self.file_path = file_path
        self.pattern = pattern

    def update(self, sessions):
        with io.open(self.file_path, "w", encoding="utf-8") as map_file:
            for id in sessions:
                map_file.write(
                    self.pattern % (id, sessions[id]["host"], sessions[id]["port"])
                )


# =============================================================================
# Resource manager
# =============================================================================


class ResourceManager(object):
    """
    Class that provides methods to keep track on available resources (host/port)
    """

    def __init__(self, resourceList):
        self.resources = {}
        for resource in resourceList:
            host = resource["host"]
            portList = list(
                range(resource["port_range"][0], resource["port_range"][1] + 1)
            )
            if host in self.resources:
                self.resources[host]["available"].extend(portList)
            else:
                self.resources[host] = {"available": portList, "used": []}

    def getNextResource(self):
        """
        Return a (host, port) pair if any available otherwise will return None
        """
        # find host with max availibility
        winner = None
        availibilityCount = 0
        for host in self.resources:
            if availibilityCount < len(self.resources[host]["available"]):
                availibilityCount = len(self.resources[host]["available"])
                winner = host

        if winner:
            port = self.resources[winner]["available"].pop()
            self.resources[winner]["used"].append(port)
            return (winner, port)

        return (None, None)

    def freeResource(self, host, port):
        """
        Free a previously reserved resource
        """
        if host in self.resources and port in self.resources[host]["used"]:
            self.resources[host]["used"].remove(port)
            self.resources[host]["available"].append(port)


# =============================================================================
# Process manager
# =============================================================================


class ProcessManager(object):
    def __init__(self, configuration):
        self.config = configuration
        self.log_dir = configuration["configuration"]["log_dir"]
        self.processes = {}

    def __del__(self):
        for id in self.processes:
            self.processes[id].terminate()

    def _getLogFilePath(self, id):
        return "%s%s%s.txt" % (self.log_dir, os.sep, id)

    def startProcess(self, session):
        proc = None

        # Create output log file
        logFilePath = self._getLogFilePath(session["id"])
        with io.open(logFilePath, mode="a+", buffering=1, encoding="utf-8") as log_file:
            try:
                proc = subprocess.Popen(
                    session["cmd"], stdout=log_file, stderr=log_file
                )
                self.processes[session["id"]] = proc
            except:
                logger.error("The command line failed")
                logger.error(" ".join(map(str, session["cmd"])))
                return None

        return proc

    def stopProcess(self, id):
        proc = self.processes[id]
        del self.processes[id]
        try:
            proc.terminate()
        except:
            pass  # we tried

    def listEndedProcess(self):
        session_to_release = []
        for id in self.processes:
            if self.processes[id].poll() is not None:
                session_to_release.append(id)
        return session_to_release

    def isRunning(self, id):
        return self.processes[id].poll() is None

    # ========================================================================
    # Look for ready line in process output. Return True if found, False
    # otherwise. If no ready_line is configured and process is running return
    # False. This will then rely on the timout time.
    # ========================================================================

    def isReady(self, session, count=0):
        id = session["id"]

        # The process has to be running to be ready!
        if not self.isRunning(id) and count < 60:
            return False

        # Give up after 60 seconds if still not running
        if not self.isRunning(id):
            return True

        application = self.config["apps"][session["application"]]
        ready_line = application.get("ready_line", None)

        # If no ready_line is configured and the process is running then thats
        # enough.
        if not ready_line:
            return False

        ready = False

        # Check the output for ready_line
        logFilePath = self._getLogFilePath(session["id"])
        with io.open(logFilePath, "r", 1, encoding="utf-8") as log_file:
            for line in log_file.readlines():
                if ready_line in line:
                    ready = True
                    break

        return ready


# =============================================================================
# Parse config file
# =============================================================================


def parseConfig(options):
    # Read values from the configuration file
    try:
        config_comments = remove_comments(
            io.open(options.config[0], encoding="utf-8").read()
        )
        config = json.loads(config_comments)
    except:
        message = "ERROR: Unable to read config file.\n"
        message += str(sys.exc_info()[1]) + "\n" + str(sys.exc_info()[2])
        print(message)
        print(sample_config_file)
        sys.exit(2)

    expected_keys = ["configuration", "apps", "properties", "resources"]
    if not validateKeySet(config, expected_keys, "Config file"):
        print(sample_config_file)
        sys.exit(2)

    expected_keys = [
        "endpoint",
        "host",
        "port",
        "proxy_file",
        "sessionURL",
        "timeout",
        "log_dir",
        "fields",
    ]
    if not validateKeySet(config["configuration"], expected_keys, "file.configuration"):
        print(sample_config_file)
        sys.exit(2)

    if not "content" in config["configuration"]:
        config["configuration"]["content"] = ""
    if not "sanitize" in config["configuration"]:
        config["configuration"]["sanitize"] = {}

    return config


# =============================================================================
# Setup default arguments to be parsed
#   -d, --debug
#   -t, --proxyFileType  Type of proxy file (txt, dbm)
# =============================================================================


def add_arguments(parser):
    parser.add_argument(
        "config", type=str, nargs=1, help="configuration file for the launcher"
    )
    parser.add_argument(
        "-d", "--debug", help="log debugging messages to stdout", action="store_true"
    )
    parser.add_argument(
        "--backend", help="Server implementation to use (aiohttp)", default="aiohttp"
    )

    return parser


# =============================================================================
# Parse arguments
# =============================================================================


def start(argv=None, description="wslink Web Launcher"):
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=sample_config_file,
    )
    add_arguments(parser)
    args = parser.parse_args(argv)
    config = parseConfig(args)
    backends.launcher_start(args, config, backend=args.backend)


# Used for backward compatibility


def startWebServer(args, config, backend="aiohttp"):
    return backends.launcher_start(args, config, backend=backend)


# =============================================================================
# Main
# =============================================================================

if __name__ == "__main__":
    start()



r"""
Wslink allows easy, bi-directional communication between a python server and a
javascript client over a websocket.

wslink.server creates the python server
wslink.websocket handles the communication
"""

import asyncio
import functools

from .uri import checkURI

__license__ = "BSD-3-Clause"


def register(uri):
    """
    Decorator for RPC procedure endpoints.
    """

    def decorate(f):
        # called once when method is decorated, because we return 'f'.
        assert callable(f)
        if not hasattr(f, "_wslinkuris"):
            f._wslinkuris = []
        f._wslinkuris.append({"uri": checkURI(uri)})
        return f

    return decorate


#############################################################################
#
#                         scheduling methods
#
# Allow scheduling both callbacks and coroutines from both sync and async
# methods.
#
#############################################################################


def schedule_callback(delay, callback, *args, **kwargs):
    """
    Schedule callback (which is passed args and kwargs) to be called on
    running event loop after delay seconds (can be floating point).  Returns
    asyncio.TimerHandle on which cancel() can be called to cancel the
    eventual invocation of the callback.
    """
    # Using "asyncio.get_running_loop()" requires the event loop to be running
    # already, so we use "asyncio.get_event_loop()" here so that we can support
    # scheduling tasks before the server is started.
    loop = asyncio.get_event_loop()
    return loop.call_later(delay, functools.partial(callback, *args, **kwargs))


def schedule_coroutine(delay, coro_func, *args, done_callback=None, **kwargs):
    """
    Creates a coroutine out of the provided coroutine function coro_func and
    the provided args and kwargs, then schedules the coroutine to be called
    on the running event loop after delay seconds (delay can be float or int).

    Returns asyncio.Task on which cancel() can be called to cancel the running
    of the coroutine.

    The coro_func parameter should not be a coroutine, but rather a coroutine
    function (a function defined with async).  The reason for this is we want
    to defer creation of the actual coroutine until we're ready to schedule it
    with "ensure_future".  Otherwise every time we cancel a TimerTask
    returned by "call_later", python prints "RuntimeWarning: coroutine
    '<coro-name>' was never awaited".
    """
    # See method above for comment on "get_event_loop()" vs "get_running_loop()".
    loop = asyncio.get_event_loop()
    coro_partial = functools.partial(coro_func, *args, **kwargs)
    if done_callback is not None:
        return loop.call_later(
            delay,
            lambda: asyncio.ensure_future(coro_partial()).add_done_callback(
                done_callback
            ),
        )
    return loop.call_later(delay, lambda: asyncio.ensure_future(coro_partial()))



from . import schedule_coroutine

# =============================================================================
# singleton publish manager


class PublishManager(object):
    def __init__(self):
        self.protocols = []
        self.publishCount = 0

    def registerProtocol(self, protocol):
        self.protocols.append(protocol)

    def unregisterProtocol(self, protocol):
        if protocol in self.protocols:
            self.protocols.remove(protocol)

    def addAttachment(self, payload):
        """Deprecated method, keeping it to avoid breaking compatibility
        Now that we use msgpack to pack/unpack messages,
        We can have binary data directly in the object itself,
        without needing to transfer it separately from the rest."""
        return payload

    def publish(self, topic, data, client_id=None, skip_last_active_client=False):
        for protocol in self.protocols:
            # The client is unknown - we send to any client who is subscribed to the topic
            rpcid = "publish:{0}:{1}".format(topic, self.publishCount)
            protocol.network_monitor.on_enter()
            schedule_coroutine(
                0,
                protocol.sendWrappedMessage,
                rpcid,
                data,
                client_id=client_id,
                skip_last_active_client=skip_last_active_client,
                # for schedule_coroutine call
                done_callback=protocol.network_monitor.on_exit,
            )


# singleton, used by all instances of WslinkWebSocketServerProtocol
publishManager = PublishManager()



import ssl


def load_ssl_context(cert_file, pkey_file):
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    context.load_cert_chain(cert_file, pkey_file)
    return context


def save_ssl_files(cert, pkey):
    import atexit
    import os
    import tempfile
    from cryptography.hazmat.primitives import serialization

    cert_handle, cert_file = tempfile.mkstemp()
    pkey_handle, pkey_file = tempfile.mkstemp()
    atexit.register(os.remove, pkey_file)
    atexit.register(os.remove, cert_file)

    os.write(cert_handle, cert.public_bytes(serialization.Encoding.PEM))
    os.write(
        pkey_handle,
        pkey.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        ),
    )

    os.close(cert_handle)
    os.close(pkey_handle)
    return cert_file, pkey_file


def generate_ssl_pair(host):
    try:
        from cryptography import x509
        from cryptography.x509.oid import NameOID
        from cryptography.hazmat.primitives import hashes
        from cryptography.hazmat.primitives.asymmetric import rsa
        import datetime
    except ImportError:
        raise TypeError(
            "Using ad-hoc certificates requires the cryptography library."
        ) from None
    cn = f"*.{host}/CN={host}"
    pkey = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    subject = x509.Name(
        [
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Dummy Certificate"),
            x509.NameAttribute(NameOID.COMMON_NAME, cn),
        ]
    )
    one_day = datetime.timedelta(1, 0, 0)
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(subject)
        .public_key(pkey.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.today() - one_day)
        .not_valid_after(datetime.datetime.today() + (one_day * 365))
        .add_extension(x509.ExtendedKeyUsage([x509.OID_SERVER_AUTH]), critical=False)
        .add_extension(x509.SubjectAlternativeName([x509.DNSName(cn)]), critical=False)
        .sign(private_key=pkey, algorithm=hashes.SHA256())
    )
    return save_ssl_files(cert, pkey)



import asyncio
from functools import partial
from wslink.backends.generic.core import GenericServer
from IPython.core.getipython import get_ipython


class EventEmitter:
    def __init__(self):
        self._listeners = {}

    def clear(self):
        self._listeners = {}

    def emit(self, event, *args, **kwargs):
        listeners = self._listeners.get(event)
        if listeners is None:
            return

        loop = asyncio.get_running_loop()
        coroutine_run = (
            loop.create_task if (loop and loop.is_running()) else asyncio.run
        )

        for listener in listeners:
            if asyncio.iscoroutinefunction(listener):
                coroutine_run(listener(*args, **kwargs))
            else:
                listener(*args, **kwargs)

    def add_event_listener(self, event, listener):
        listeners = self._listeners.get(event)
        if listeners is None:
            listeners = set()
            self._listeners[event] = listeners

        listeners.add(listener)

    def remove_event_listener(self, event, listener):
        listeners = self._listeners.get(event)
        if listeners is None:
            return

        if listener in listeners:
            listeners.remove(listener)


class WsJupyterComm(EventEmitter):
    def __init__(self, kernel=None):
        super().__init__()
        self.comm = None
        self.kernel = get_ipython().kernel if kernel is None else kernel
        self.kernel.comm_manager.register_target("wslink_comm", self.on_open)

    def send(self, data, buffers):
        if self.comm is not None:
            self.comm.send(data=data, buffers=buffers)

    def on_message(self, msg):
        self.emit("message", msg["content"]["data"], msg["buffers"])

    def on_close(self, msg):
        self.comm = None

    def on_open(self, comm, msg):
        self.comm = comm
        comm.on_msg(self.on_message)
        comm.on_close(self.on_close)


JUPYTER_COMM = None


def get_jupyter_comm(kernel=None):
    global JUPYTER_COMM
    if JUPYTER_COMM is None:
        JUPYTER_COMM = WsJupyterComm(kernel)

    return JUPYTER_COMM


class GenericMessage:
    def __init__(self, data):
        self.data = data


class JupyterGenericServer(GenericServer):
    def __init__(self, server_config):
        super().__init__(server_config)

        self.trame_comm = get_jupyter_comm()
        self._endpoint = self[self.ws_endpoints[0]]
        self._name = self._endpoint.serverProtocol.server.name
        self._connections = {}
        self.trame_comm.add_event_listener("message", self.on_msg_from_comm)

    async def on_msg_from_server(self, client_id, binary, content):
        buffers = []
        data = {"server": self._name, "client": client_id}

        if binary:
            buffers.append(content)
        else:
            data["payload"] = content

        self.trame_comm.send(data, buffers)

    async def on_msg_from_comm(self, data, buffers):
        server_name = data["server"]
        client_id = data["client"]

        if server_name != self._name:
            return

        connection = self._connections.get(client_id, None)

        if connection is None:
            connection = await self._endpoint.connect()
            connection.on_message(partial(self.on_msg_from_server, client_id))
            self._connections[client_id] = connection

        is_binary = len(buffers) > 0

        message = None

        if is_binary:
            message = GenericMessage(buffers[0])
        else:
            message = GenericMessage(data["payload"])

        await connection.send(is_binary, message)


def startWebServer(*args, **kwargs):
    raise NotImplementedError("Generic backend does not provide a launcher")


def create_webserver(server_config):
    jupyter_generic_server = JupyterGenericServer(server_config)
    return jupyter_generic_server
